from pokie.contrib.auth.constants import SVC_USER, SVC_ACL, SVC_AUTH
from pokie.contrib.mail.constants import SVC_MESSAGE_TEMPLATE, SVC_MESSAGE_QUEUE
from pokie.core import BaseModule


class Module(BaseModule):
    name = "mail"
    description = "Mail template module"

    services = {
        SVC_MESSAGE_TEMPLATE: "pokie.contrib.mail.service.MessageTemplateService",
        SVC_MESSAGE_QUEUE: "pokie.contrib.mail.service.MessageQueueService",
    }

    cmd = {
        "mail:purge": "pokie.contrib.mail.cli.PurgeQueueCmd",
        "mail:run": "pokie.contrib.mail.cli.RunQueueCmd",
    }

    jobs = [
        "pokie.contrib.mail.job.MailQueueJob",
    ]



from argparse import ArgumentParser

from pokie.constants import DI_SERVICES, DI_APP
from pokie.contrib.base.constants import SVC_FIXTURE
from pokie.contrib.base.dto import FixtureRecord
from pokie.contrib.base.service.fixture import FixtureService
from pokie.contrib.mail.constants import SVC_MESSAGE_QUEUE
from pokie.contrib.mail.job import MailQueueJob
from pokie.contrib.mail.service import MessageQueueService
from pokie.core import CliCommand


class QueueCmd(CliCommand):
    @property
    def svc_queue(self) -> MessageQueueService:
        return self.get_di().get(DI_SERVICES).get(SVC_MESSAGE_QUEUE)


class RunQueueCmd(CliCommand):
    description = "run mail queue"

    def run(self, args) -> bool:
        self.tty.write("Running mail queue...")
        job_runner = MailQueueJob(self.get_di())
        job_runner.run(self.get_di())
        self.tty.write("Done!")

        return True


class PurgeQueueCmd(QueueCmd):
    description = "purge mail queue (delete all entries)"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "--yes",
            default=False,
            action="store_true",
            help="Confirm purge",
        )

    def run(self, args) -> bool:
        if not args.yes:
            self.tty.error("missing --yes argument, aborting...")
            return False

        self.tty.write("Removing all mail queue entries...")
        self.svc_queue.purge()
        self.tty.write("Purging finished.")
        return True



# predefined method names and route rule expansions
controller_action_map = {
    "list": ["/{slug}", ["GET"], "_list"],
    "show": ["/{slug}/<{type}:id>", ["GET"], "_show"],
    "create": ["/{slug}", ["POST"], "_create"],
    "update": ["/{slug}/<{type}:id>", ["PUT", "PATCH"], "_update"],
    "delete": ["/{slug}/<{type}:id>", ["DELETE"], "_delete"],
}

# predefined resource methods and route rule expansions
resource_action_map = {
    # method_name: [rule, rule, ...]
    "get": [["/{slug}", ["GET"], "_list"], ["/{slug}/<{type}:id>", ["GET"], "_show"]],
    "post": [["/{slug}", ["POST"], "_create"]],
    "put": [["/{slug}/<{type}:id>", ["PUT", "PATCH"], "_update"]],
    "delete": [["/{slug}/<{type}:id>", ["DELETE"], "_delete"]],
}


def route_controller(app, slug: str, cls, id_type: str = "int"):
    """
    Register default routes for a controller class

    :param app: Flask app or blueprint
    :param slug: route base prefix
    :param cls: class to map
    :param id_type: optional datatype for id
    :return:
    """
    name = cls.__name__
    for method_name, opts in controller_action_map.items():
        route, methods, suffix = opts
        if callable(getattr(cls, method_name, None)):
            app.add_url_rule(
                route.format(slug=slug, type=id_type),
                methods=methods,
                view_func=cls.view_method(method_name),
            )


def route_resource(app, slug, cls, id_type: str = "int"):
    """
    Register default routes for a resource class

    :param app: Flask app or blueprint
    :param slug: route base prefix
    :param cls: class to map
    :param id_type: optional datatype for id
    :return:
    """
    name = ".".join([cls.__module__, cls.__name__]).replace(".", "_")
    for view_name, routes in resource_action_map.items():
        for item in routes:
            route, methods, suffix = item
            if getattr(cls, view_name, None) is not None:
                app.add_url_rule(
                    route.format(slug=slug, type=id_type),
                    methods=methods,
                    view_func=cls.as_view("{}{}".format(name, suffix)),
                )



from typing import List, Optional

from rick.mixin.injectable import Injectable

from pokie.contrib.auth.dto import AclRole, AclResource
from pokie.contrib.auth.repository.acl import AclRoleRepository, AclResourceRepository
from pokie.constants import DI_DB


class AclService(Injectable):
    def get_user_acl_info(self, id_user: int) -> dict:
        return {
            "roles": self.list_user_role_id(id_user),
            "resources": self.list_user_resource_id(id_user),
        }

    def get_user_roles(self, id_user: int) -> List[AclRole]:
        return self.role_repository.find_user_roles(id_user)

    def list_user_resource_id(self, id_user: int) -> List[str]:
        result = []
        for record in self.resource_repository.find_user_resources(id_user):
            result.append(record.id)
        return result

    def list_user_role_id(self, id_user: int) -> List[str]:
        result = []
        for record in self.role_repository.find_user_roles(id_user):
            result.append(record.id)
        return result

    def list_roles(self) -> List[AclRole]:
        return self.role_repository.fetch_all_ordered(AclRole.id)

    def list_resources(self) -> List[AclResource]:
        return self.resource_repository.fetch_all_ordered(AclResource.id)

    def list_role_resources(self, id_role: int) -> List[AclResource]:
        return self.resource_repository.find_by_role(id_role)

    def get_role(self, id_role: int) -> Optional[AclRole]:
        return self.role_repository.fetch_pk(id_role)

    def get_resource(self, id_resource: str) -> Optional[AclResource]:
        return self.resource_repository.fetch_pk(id_resource)

    def add_role(self, description: str) -> int:
        return self.role_repository.insert_pk(AclRole(description=description))

    def add_resource(self, id_resource: str, description: str) -> int:
        return self.resource_repository.insert_pk(
            AclResource(id=id_resource, description=description)
        )

    def add_role_resource(self, id_role: int, id_resource: int):
        return self.role_repository.add_role_resource(id_role, id_resource)

    def add_user_role(self, id_user: int, id_role: int):
        self.role_repository.add_user_role(id_user, id_role)

    def remove_user_role(self, id_user: int, id_role: int):
        self.role_repository.remove_user_role(id_user, id_role)

    def remove_role(self, id_role: int):
        return self.role_repository.delete_pk(id_role)

    def remove_resource(self, id_resource: int):
        return self.resource_repository.delete_pk(id_resource)

    def can_remove_role(self, id_role: int) -> bool:
        return self.role_repository.can_remove(id_role)

    def truncate_role(self, id_role: int):
        return self.role_repository.truncate(id_role)

    def can_remove_resource(self, id_resource: int) -> bool:
        return self.resource_repository.can_remove(id_resource)

    def truncate_resource(self, id_resource: int):
        return self.resource_repository.truncate(id_resource)

    def remove_role_resource(self, id_role: int, id_resource: int):
        return self.role_repository.remove_role_resource(id_role, id_resource)

    @property
    def role_repository(self):
        return AclRoleRepository(self.get_di().get(DI_DB))

    @property
    def resource_repository(self):
        return AclResourceRepository(self.get_di().get(DI_DB))



from .base import ListCmd, HelpCmd, RunServerCmd, VersionCmd
from .db import DbInitCmd, DbCheckCmd, DbUpdateCmd
from .job import JobRunCmd, JobListCmd
from .db_codegen import GenDtoCmd, GenRequestRecordCmd
from .tpl_codegen import ModuleGenCmd
from .fixture import RunFixtureCmd, CheckFixtureCmd
from .pytest import PyTestCmd



import json
import logging
from typing import Any, Optional, Callable
from flask import request
from flask.views import MethodView
from flask.typing import ResponseReturnValue
from flask import current_app
from flask_login import current_user
from rick.form import RequestRecord

from .response import JsonResponse
from pokie.constants import (
    HTTP_OK,
    HTTP_BADREQ,
    HTTP_INTERNAL_ERROR,
    HTTP_NOAUTH,
    HTTP_FORBIDDEN,
    DI_SERVICES,
)


class PokieView(MethodView):
    # allowed HTTP methods
    allow_methods = ["get", "post", "put", "patch", "delete", "head"]

    # optional RequestRecord class for request body unmarshall
    # see self.request below
    request_class = None  # type: RequestRecord

    # default response class
    response_class = JsonResponse

    # default error message
    msg_error_default = "request failed"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.di = current_app.di
        self.logger = current_app.logger

        # methods where automatic body deserialization is attempted
        #
        # names must be lowercase
        self.methods_unmarshall_request = ["post", "put", "patch"]

        # automatic request de-serialization
        #
        # if request.method is in self.methods_unmarshall_request, we will attempt to de-serialize the body
        # and validate it using the request_class data type
        self.request = None

        # pre-dispatch hooks
        # these are called before performing the actual dispatch, with the folloing interface:
        # (method:str, *args: t.Any, **kwargs: t.Any) -> Optional[ResponseReturnValue]
        # if they generate a response other than None, dispatch is aborted and that response is used
        # Note: they are not chained - eg. the first one to return something aborts dispatching
        self.dispatch_hooks = [
            "_hook_request",
        ]

        # optional override of internal options
        for name, value in kwargs.items():
            attr = getattr(self, name, None)
            if attr is not None and not callable(attr):
                setattr(self, name, value)

    def _hook_request(
        self, method: str, *args: Any, **kwargs: Any
    ) -> Optional[ResponseReturnValue]:
        """
        Dispatch hook: de-serialize request
        If body is not json, will use flask form data

        :param method: method name in lowercase
        :param args:
        :param kwargs:
        :return: ResponseReturnValue or None
        """
        if self.request_class is None:
            return None

        if method not in self.methods_unmarshall_request:
            return None

        # unmarshall
        data = None
        if request.is_json:
            if request.content_length is not None:
                data = request.json
        else:
            # form-data
            data = request.form.items()

        if data is None or len(data) == 0:
            return self.error("empty body")

        # validate request body
        self.request = self.request_class()
        if not isinstance(self.request, RequestRecord):
            raise ValueError(
                "_pre_request(): invalid request class; class must extend RequestRecord"
            )
        if self.request.is_valid(data):
            return None
        return self.request_error(self.request)

    def dispatch_request(self, *args: Any, **kwargs: Any) -> ResponseReturnValue:
        """
        Main request dispatcher

        Extends the Flask implementation by providing pre-dispatch hooks

        :param args:
        :param kwargs:
        :return: ResponseReturnValue
        """
        method = request.method.lower()
        handler = getattr(self, method, None)

        if method not in self.allow_methods:
            return self.exception_handler(None)

        # If the request method is HEAD and we don't have a handler for it
        # retry with GET.
        if handler is None and method == "head":
            handler = getattr(self, "get", None)
        # support for named views
        if "_action_method_" in kwargs.keys():
            handler = getattr(self, kwargs["_action_method_"], None)
            del kwargs["_action_method_"]

        assert handler is not None, "Cannot resolve handler method for dispatch"

        try:
            # run pre-dispatch hooks
            for name in self.dispatch_hooks:
                hook = getattr(self, name, None)
                assert hook is not None, f"non-existing dispatch hook {name!r}"
                pre = hook(method, *args, **kwargs)
                if pre is not None:
                    return pre

            return current_app.ensure_sync(handler)(*args, **kwargs)
        except Exception as e:
            return self.exception_handler(e)

    @classmethod
    def view_method(
        cls, action_method: str, name=None, *class_args: Any, **class_kwargs: Any
    ) -> Callable:
        """
        Variant of Flask's as_view that supports custom handlers for actions
        :param action_method: method to be called on dispatch 
        :param name: optional route name
        :param class_args: 
        :param class_kwargs: 
        :return: Callable
        """ """
        """
        if name is None:
            name = ".".join([cls.__module__, cls.__name__, action_method]).replace(
                ".", "_"
            )

        def view(*args: Any, **kwargs: Any) -> ResponseReturnValue:
            self = view.view_class(*class_args, **class_kwargs)  # type: ignore
            # add the action method to the dispatch arguments
            kwargs["_action_method_"] = action_method
            return current_app.ensure_sync(self.dispatch_request)(*args, **kwargs)

        if cls.decorators:
            view.__name__ = name
            view.__module__ = cls.__module__
            for decorator in cls.decorators:
                view = decorator(view)

        # We attach the view class to the view function for two reasons:
        # first of all it allows us to easily figure out what class-based
        # view this thing came from, secondly it's also used for instantiating
        # the view class so you can actually replace it with something else
        # for testing purposes and debugging.
        view.view_class = cls  # type: ignore
        view.__name__ = name
        view.__doc__ = cls.__doc__
        view.__module__ = cls.__module__
        view.methods = cls.methods  # type: ignore
        view.provide_automatic_options = cls.provide_automatic_options  # type: ignore
        return view

    def exception_handler(self, e) -> ResponseReturnValue:
        """
        Generic exception handler for dispatch
        :param e:
        :return:
        """
        if e is not None:
            self.logger.exception(e)
        return self.error("bad request")

    def success(self, data=None, code: int = HTTP_OK):
        """
        Returns a success response with optional data payload
        :param data: optional response data
        :param code: int
        :return: Response
        """
        cls = self.response_class(data=data, success=True, code=code)
        return cls.assemble(current_app)

    def success_message(self, message: str):
        """
        Returns a success response with a string message
        :param message: str
        :return: Response
        """
        return self.success({"message": message})

    def error(self, message=None, code=HTTP_BADREQ):
        """
        Returns an error response with an optional string message
        if no message is provided, defaults to PokieView.msg_error_default
        :param message: str
        :param code: int
        :return: Response
        """

        if message is None:
            message = self.msg_error_default

        cls = self.response_class(error={"message": message}, success=False, code=code)
        return cls.assemble(current_app)

    def request_error(self, request_object: RequestRecord, code=HTTP_BADREQ):
        """
        Returns a RequestRecord error response

        :param request_object: RequestRecord
        :param code: int
        :return: Response
        """
        error = {
            "message": self.msg_error_default,
            "formError": request_object.get_errors(),
        }
        cls = self.response_class(error=error, success=False, code=code)
        return cls.assemble(current_app)

    def empty_body(self):
        """
        Error: empty body
        :return: Request
        """
        return self.error("empty body", code=HTTP_BADREQ)

    def not_found(self):
        """
        Error: not found
        :return: Request
        """
        return self.error("record not found", code=HTTP_BADREQ)

    def forbidden(self):
        """
        Error: access denied
        :return: Request
        """
        return self.error("access denied", code=HTTP_FORBIDDEN)

    def denied(self):
        """
        Error: access denied
        :return: Request
        """
        return self.error("access denied", code=HTTP_NOAUTH)

    def get_service(self, service_name):
        return self.di.get(DI_SERVICES).get(service_name)


class PokieAuthView(PokieView):
    # list of acls to check for current user
    # if list is empty, no acl control is used
    acl = []
    user = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # auth first
        self.dispatch_hooks = ["_hook_auth"] + self.dispatch_hooks
        self.user = current_user

    def _hook_auth(
        self, method: str, *args: Any, **kwargs: Any
    ) -> Optional[ResponseReturnValue]:
        if not current_user.is_authenticated:
            return self.denied()

        for acl in self.acl:
            if not self.user.can_access(acl):
                return self.forbidden()

        return None



from typing import Optional

from rick_db.conn import Connection
from rick_db.util.pg.records import ColumnRecord
from rick_db.util.pg import PgInfo

from pokie.codegen.spec import TableSpec, FieldSpec


class PgTableSpec:
    def __init__(self, conn: Connection):
        self.db = conn
        self.mgr = PgInfo(conn)
        self._tables = {}
        self._indexes = {}

    def manager(self):
        return self.mgr

    def get_pk(self, table, schema) -> Optional[str]:
        for f in self.mgr.list_table_indexes(table, schema):
            if f.primary:
                return f.field
        return None

    def get_fields(self, table, schema) -> dict:
        result = {}
        for c in self.mgr.list_table_columns(table, schema):
            result[c.column] = c
        return result

    def is_serial(self, table, field, schema) -> bool:
        namespec = "{}.{}".format(schema, table)
        sql = "SELECT pg_get_serial_sequence(%s, %s)"
        with self.db.cursor() as c:
            return len(c.exec(sql, (namespec, field))) > 0

    def get_fk(self, table, schema) -> dict:
        result = {}
        for r in self.mgr.list_table_foreign_keys(table, schema):
            result[r.column] = r
        return result

    def spec_bpchar(self, f: ColumnRecord) -> dict:
        if f.maxlen is not None:
            return {"maxlen": f.maxlen}
        return {}

    def spec_varchar(self, f: ColumnRecord) -> dict:
        if f.maxlen is not None:
            return {"maxlen": f.maxlen}
        return {}

    def spec_numeric(self, f: ColumnRecord) -> dict:
        return {
            "precision": f.numeric_precision,
            "cardinal": f.numeric_precision_cardinal,
        }

    def table_spec(self, table, schema: str = None) -> TableSpec:
        """
        Generate a table spec for the given table
        :param table:
        :param schema:
        :return: TableSpec object
        """
        if schema is None:
            schema = PgInfo.SCHEMA_DEFAULT

        pk = self.get_pk(table, schema)
        fks = self.get_fk(table, schema)
        fields = self.get_fields(table, schema)
        identity = None
        pk_auto = False

        for name, f in fields.items():
            if f.is_identity == "YES":
                identity = name
                break

        # primary key may not exist as key, but table may have an identity column
        # if we find an always generated identity column, we'll use that
        if pk is None and identity is not None:
            pk = identity
            pk_auto = True

        if pk is not None:
            if pk not in fields.keys():
                raise RuntimeError(
                    "Primary key '{}' does not exist in table field list for table {}.{}".format(
                        pk, schema, table
                    )
                )

            if not pk_auto:
                # pk_auto is true if pk is serial or if pk is an identity column
                pk_auto = self.is_serial(table, pk, schema) or str(pk) == str(identity)

        spec = TableSpec(table=table, schema=schema, pk=pk, fields=[])
        for name, f in fields.items():
            is_pk = pk == f.column
            auto = is_pk and pk_auto

            spec_formatter = getattr(self, "spec_" + f.udt_name, None)
            type_spec = {}
            if callable(spec_formatter):
                type_spec = spec_formatter(f)

            field = FieldSpec(
                name=f.column,
                pk=is_pk,
                auto=auto,
                nullable=f.is_nullable == "YES",
                fk=False,
                fk_table=None,
                fk_schema=None,
                fk_column=None,
                dtype=f.udt_name,
                dtype_spec=type_spec,
            )

            if f.column in fks.keys():
                field.fk = True
                field.fk_table = fks[f.column].foreign_table
                field.fk_schema = fks[f.column].foreign_schema
                field.fk_column = fks[f.column].foreign_column

            spec.fields.append(field)

        return spec



from rick_db import fieldmapper


@fieldmapper(tablename="settings", pk="id_settings")
class SettingsRecord:
    id = "id_settings"
    module = "module"
    key = "key"
    value = "value"


@fieldmapper(tablename="_fixture", pk="id_fixture")
class FixtureRecord:
    id = "id_fixture"
    applied = "applied"
    name = "name"



class RestServiceMixin:
    def get(self, id_record):
        return self.repository.fetch_pk(id_record)

    def delete(self, id_record):
        return self.repository.delete_pk(id_record)

    def insert(self, record):
        return self.repository.insert_pk(record)

    def update(self, id_record, record):
        return self.repository.update(record, id_record)

    def exists(self, id_record):
        pass

    def list(
        self,
        search_fields: list,
        search_text: str = None,
        match_fields: dict = None,
        limit: int = None,
        offset: int = None,
        sort_fields: dict = None,
    ):
        pass



from rick.base import Di
from rick.mixin import Injectable

from pokie.constants import HTTP_NOT_FOUND, HTTP_INTERNAL_ERROR, DI_FLASK
from .response import JsonResponse


class HttpErrorHandler(Injectable):
    ERROR_404 = "404 Not Found: The requested URL was not found on the server."
    ERROR_500 = "500 Internal Server Error"

    def __init__(self, di: Di):
        super().__init__(di)
        _app = di.get(DI_FLASK)

        def wrapper_404(e):
            return self.error_404(_app, e)

        def wrapper_500(e):
            return self.error_500(_app, e)

        # register global error handler
        _app.register_error_handler(404, wrapper_404)
        _app.register_error_handler(500, wrapper_500)

    def error_404(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_404}, success=False, code=HTTP_NOT_FOUND
        )
        return r.assemble(_app)

    def error_500(self, _app, e):
        r = self.response(
            error={"message": self.ERROR_500}, success=False, code=HTTP_INTERNAL_ERROR
        )
        return r.assemble(_app)

    def response(self, **kwargs):
        return JsonResponse(**kwargs)



import os
from pathlib import Path
from typing import Optional


class TemplateProcessor:
    def __init__(self, template_paths: list = None):
        if template_paths is None:
            self.template_paths = template_paths
        self.template_paths = template_paths

    def get_template_path(self, name: str) -> Optional[Path]:
        for p in self.template_paths:
            d = Path(p)
            if d.exists() and d.is_dir():
                for item in d.iterdir():
                    if item.is_dir() and d.name == name:
                        return item
        return None

    def process(self, src_path: Path, dest_path: Path, vars: dict, tty=None):
        if not src_path.exists():
            raise ValueError(
                "TemplateProcessor::process() invalid source directory '{}'".format(
                    str(src_path)
                )
            )
        if not src_path.is_dir():
            raise ValueError(
                "TemplateProcessor::process() path '{}' is not a directory".format(
                    str(src_path)
                )
            )
        if dest_path.exists() and not dest_path.is_dir():
            raise ValueError(
                "TemplateProcessor::process() path '{}' is not a directory".format(
                    str(src_path)
                )
            )

        if not dest_path.exists():
            dest_path.mkdir()
        self._process_dir(src_path, dest_path, vars, tty)

    def _process_dir(self, src: Path, dest: Path, vars: dict, tty=None):
        for f in src.iterdir():
            if f.is_file():
                contents = self.read_tpl(f)
                for var, replacement in vars.items():
                    contents = contents.replace(var, replacement)
                fname = f.name.replace(".tpl", "")
                dest_file = dest / Path(fname)
                with open(dest_file, "w") as outfile:
                    outfile.write(contents)
                if tty:
                    tty.write("created '{}' file".format(str(dest_file)))
            elif f.is_dir():
                new_dest = dest / Path(f.name)
                new_dest.mkdir()
                if tty:
                    tty.write("created '{}' directory".format(str(new_dest)))
                self._process_dir(f, new_dest, vars, tty)
            else:
                raise RuntimeError(
                    "something went wrong processing path '{}'".format(f)
                )

    def read_tpl(self, src_file: Path) -> str:
        with open(src_file, "r") as f:
            return f.read()



from argparse import ArgumentParser
from rick.form import RequestRecord, field

from pokie.contrib.auth.constants import SVC_USER, SVC_ACL
from pokie.constants import DI_SERVICES
from pokie.contrib.auth.service import UserService, AclService
from pokie.core import CliCommand


class DescriptionRequest(RequestRecord):
    fields = {
        "description": field(
            validators="required|minlen:1|maxlen:200", error="invalid description"
        )
    }


class IdDescriptionRequest(RequestRecord):
    fields = {
        "id_resource": field(
            validators="required|minlen:1|maxlen:200", error="invalid resource id"
        ),
        "description": field(
            validators="required|minlen:1|maxlen:200", error="invalid description"
        ),
    }


class UserRoleRequest(RequestRecord):
    fields = {
        "username": field(
            validators="required|pk:user,username", error="invalid username"
        ),
        "id_role": field(
            validators="required|numeric|pk:acl_role", error="invalid role id"
        ),
    }


class RoleResourceRequest(RequestRecord):
    fields = {
        "id_role": field(
            validators="required|numeric|pk:acl_role", error="invalid role id"
        ),
        "id_resource": field(
            validators="required|pk:acl_resource", error="invalid resource id"
        ),
    }


class AclCommand(CliCommand):
    @property
    def svc_user(self) -> UserService:
        return self.get_di().get(DI_SERVICES).get(SVC_USER)

    @property
    def svc_acl(self) -> AclService:
        return self.get_di().get(DI_SERVICES).get(SVC_ACL)


class AclRoleListCmd(AclCommand):
    description = "list acl roles"

    def run(self, args) -> bool:
        for record in self.svc_acl.list_roles():
            self.tty.write("{:<8}: {}".format(str(record.id), record.description))
        return True


class AclRoleCreateCmd(AclCommand):
    description = "create acl role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "description", type=str, help="Role description for new role"
        )

    def run(self, args) -> bool:
        req = DescriptionRequest()
        if not req.is_valid({"description": args.description}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        lower = args.description.lower()
        for record in self.svc_acl.list_roles():
            if record.description.lower() == lower:
                self.tty.write(
                    self.tty.colorizer.red(
                        "Error: role with same description already exists"
                    )
                )
                return False

        id = self.svc_acl.add_role(args.description)
        self.tty.write(
            "Created role #{} with description '{}'".format(str(id), args.description)
        )

        return True


class AclRoleRemoveCmd(AclCommand):
    description = "remove acl role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("id_role", type=int, help="Role id to remove")
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Force removal even if role is in use",
        )

    def run(self, args) -> bool:
        id = int(args.id_role)
        record = self.svc_acl.get_role(id)
        if record is None:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: role id #{} not found".format(str(args.id_role))
                )
            )
            return False

        if not self.svc_acl.can_remove_role(id) and args.force is False:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: role is in use; use -f to truncate records"
                )
            )
            return False

        self.svc_acl.truncate_role(id)
        self.tty.write("Role #{} removed successfully".format(id))
        return True


class AclRoleInfoCmd(AclCommand):
    description = "list resources for acl role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("id_role", type=int, help="Role id to list")

    def run(self, args) -> bool:
        id_role = int(args.id_role)
        record = self.svc_acl.get_role(id_role)
        if record is None:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: role id #{} not found".format(str(args.id_role))
                )
            )
            return False

        for resource in self.svc_acl.list_role_resources(id_role):
            self.tty.write("{:<12}: {}".format(resource.id, resource.description))
        return True


class AclResourceListCmd(AclCommand):
    description = "list acl resources"

    def run(self, args) -> bool:
        for record in self.svc_acl.list_resources():
            self.tty.write("{:<8}: {}".format(str(record.id), record.description))
        return True


class AclResourceCreateCmd(AclCommand):
    description = "create acl resource"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("id", type=str, help="Id for new resource")
        parser.add_argument("description", type=str, help="Description of new resource")

    def run(self, args) -> bool:
        req = IdDescriptionRequest()
        if not req.is_valid({"id": args.id, "description": args.description}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        id_lower = args.id.lower()
        lower = args.description.lower()
        for record in self.svc_acl.list_resources():
            if record.id.lower() == id_lower:
                self.tty.write(
                    self.tty.colorizer.red(
                        "Error: resource with specified id already exists"
                    )
                )
                return False
            if record.description.lower() == lower:
                self.tty.write(
                    self.tty.colorizer.red(
                        "Error: resource with specified description already exists"
                    )
                )
                return False

        id = self.svc_acl.add_resource(args.id, args.description)
        self.tty.write(
            "Created resource '{}' with description '{}'".format(
                str(id), args.description
            )
        )

        return True


class AclResourceLinkCmd(AclCommand):
    description = "associate an acl resource with an acl role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("id_role", type=int, help="Numeric id of existing role")
        parser.add_argument("id_resource", type=str, help="Id of existing resource")

    def run(self, args) -> bool:
        req = RoleResourceRequest()
        if not req.is_valid({"id_role": args.id_role, "id_resource": args.id_resource}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        id_role = int(args.id_role)
        self.svc_acl.add_role_resource(id_role, args.id_resource)
        self.tty.write(
            "Associated resource '{}' with role #{}".format(args.id_resource, id_role)
        )

        return True


class AclResourceUnlinkCmd(AclCommand):
    description = "removes the association between an acl role and an acl resource"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("id_role", type=int, help="Numeric id of existing role")
        parser.add_argument("id_resource", type=str, help="Id of existing resource")

    def run(self, args) -> bool:
        req = RoleResourceRequest()
        if not req.is_valid({"id_role": args.id_role, "id_resource": args.id_resource}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        id_role = int(args.id_role)
        self.svc_acl.remove_role_resource(id_role, args.id_resource)
        self.tty.write(
            "Removed association of resource '{}' with role #{}".format(
                args.id_resource, id_role
            )
        )

        return True


class AclRoleLinkCmd(AclCommand):
    description = "associate an user with an acl role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", type=str, help="Existing username")
        parser.add_argument("id_role", type=int, help="Numeric id of existing role")

    def run(self, args) -> bool:
        req = UserRoleRequest()
        if not req.is_valid({"username": args.username, "id_role": args.id_role}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        user = self.svc_user.get_by_username(args.username)
        if user is None:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: username '{}' not found".format(args.username)
                )
            )
            return False

        id_role = int(args.id_role)
        self.svc_acl.add_user_role(user.id, id_role)
        self.tty.write(
            "Associated user '{}' with role #{}".format(user.username, id_role)
        )

        return True


class AclRoleUnlinkCmd(AclCommand):
    description = "removes user association with a role"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", type=str, help="Existing username")
        parser.add_argument("id_role", type=int, help="Numeric id of existing role")

    def run(self, args) -> bool:
        req = UserRoleRequest()
        if not req.is_valid({"username": args.username, "id_role": args.id_role}):
            for k, v in req.get_errors().items():
                self.tty.write(
                    self.tty.colorizer.red("Error in {}: {}".format(k, v["*"]))
                )
            return False

        user = self.svc_user.get_by_username(args.username)
        if user is None:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: username '{}' not found".format(args.username)
                )
            )
            return False

        id_role = int(args.id_role)
        self.svc_acl.remove_user_role(user.id, id_role)
        self.tty.write(
            "Removed association of user '{}' with role #{}".format(
                user.username, id_role
            )
        )

        return True


class AclUserRoleCmd(AclCommand):
    description = "list roles associated with a given username"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("username", type=str, help="Existing username")

    def run(self, args) -> bool:
        user = self.svc_user.get_by_username(args.username)
        if user is None:
            self.tty.write(
                self.tty.colorizer.red(
                    "Error: username '{}' not found".format(args.username)
                )
            )
            return False

        for role in self.svc_acl.get_user_roles(user.id):
            self.tty.write("{:<12}: {}".format(role.id, role.description))

        return True



import uuid

from rick.resource.config import StrOrFile


class BaseConfigTemplate:
    # default HTTP Exception Handler - 404 and 500 exceptions
    HTTP_ERROR_HANDLER = 'pokie.http.HttpErrorHandler'

    # list of enabled module names
    MODULES = []
    # if true, all endpoints are authenticated by default
    USE_AUTH = True
    # Autentication plugins to use
    AUTH_PLUGINS = ["pokie.contrib.auth.plugin.DbAuthPlugin"]

    # Secret key for flask-login hashing
    AUTH_SECRET = uuid.uuid4().hex

    # cache table-related metadata (such as primary key info)
    # development should be false
    DB_CACHE_METADATA = False


class PgConfigTemplate:
    # Postgresql Configuration
    DB_NAME = "pokie"
    DB_HOST = "localhost"
    DB_PORT = 5432
    DB_USER = StrOrFile("postgres")
    DB_PASSWORD = StrOrFile("")
    DB_SSL = "True"


class RedisConfigTemplate:
    # Redis Configuration
    REDIS_HOST = "localhost"
    REDIS_PORT = 6379
    REDIS_PASSWORD = StrOrFile("")
    REDIS_DB = 0
    REDIS_SSL = "1"


class MailConfigTemplate:
    # Message channel configuration
    channels = {"0": "SMTP"}
    SMTP_HOST = "localhost"
    SMTP_PORT = 25
    SMTP_USE_TLS = False
    SMTP_USE_SSL = False
    SMTP_DEBUG = False
    SMTP_USERNAME = StrOrFile("username")
    SMTP_PASSWORD = StrOrFile("password")
    SMTP_DEFAULT_SENDER = None
    SMTP_TIMEOUT = None
    SMTP_SSL_KEYFILE = None
    SMTP_SSL_CERTFILE = None


class TestConfigTemplate:
    TEST_DB_NAME = "pokie_test"  # test database parameters
    TEST_DB_HOST = "localhost"
    TEST_DB_PORT = 5432
    TEST_DB_USER = StrOrFile("postgres")
    TEST_DB_PASSWORD = StrOrFile("")
    TEST_DB_SSL = False

    TEST_MANAGE_DB = (
        False  # if false, unit testing does not manage db creation/migration
    )
    TEST_SHARE_CTX = False  # if false, each test has a separate context
    TEST_DB_REUSE = False  # if true, database is not dropped/recreated
    TEST_SKIP_MIGRATIONS = False  # if true, migrations are not run when recreating db
    TEST_SKIP_FIXTURES = False  # if true, fixtures are not run when recreating db



from typing import List, Optional

from pokie.constants import DI_DB
from pokie.contrib.mail.dto import MessageQueueRecord
from pokie.contrib.mail.repository import MessageQueueRepository
from rick.mixin import Injectable


class MessageQueueService(Injectable):
    STATUS_QUEUED = "Q"
    STATUS_LOCKED = "L"
    STATUS_FAILED = "F"
    STATUS_SENT = "S"

    # communication channels
    CHANNEL_SMTP = 0

    def queue(self, record: MessageQueueRecord) -> str:
        return self.repo_queue.insert(record, cols=[MessageQueueRecord.id])

    def list_by_status(
        self, channel: int, status: str, limit: int = 1000
    ) -> List[MessageQueueRecord]:
        return self.repo_queue.find_by_status(channel, status, limit)

    def update_status(self, id, status):
        self.repo_queue.update_status(id, status)

    def fetch_for_processing(self, channel: int) -> Optional[MessageQueueRecord]:
        """
        Pick & lock a message for processing
        :return: MessageQueueRecord or None
        """
        return self.repo_queue.find_first_and_lock(channel, self.STATUS_QUEUED)

    def purge(self):
        """
        Removes all queued records
        :return:
        """
        return self.repo_queue.truncate()

    @property
    def repo_queue(self) -> MessageQueueRepository:
        return MessageQueueRepository(self.get_di().get(DI_DB))



from dataclasses import dataclass
from typing import Optional, List


@dataclass
class FieldSpec:
    name: str
    pk: bool
    auto: bool
    nullable: bool
    fk: bool
    fk_table: Optional[str]
    fk_schema: Optional[str]
    fk_column: Optional[str]
    dtype: str
    dtype_spec: dict


@dataclass
class TableSpec:
    table: str
    pk: Optional[str]
    schema: str
    fields: List[FieldSpec]



import signal
from rick.mixin import Injectable
from rick.base import Di


class SignalManager(Injectable):
    def __init__(self, di: Di):
        super().__init__(di)
        self.handlers = {}

    def add_handler(self, signalnum: int, handler: callable):
        assert callable(handler) is True
        assert isinstance(signalnum, int)

        if signalnum in self.handlers.keys():
            self.handlers[signalnum].append(handler)
        else:
            self.handlers[signalnum] = [handler]
            self._register_handler(signalnum)

    def _register_handler(self, signalnum: int):
        def wrap_signal(signal_no, stack_frame):
            for handler in self.handlers[signal_no]:
                handler(self.get_di(), signal_no, stack_frame)

        signal.signal(signalnum, wrap_signal)



from typing import Optional

from pokie.constants import DI_DB
from rick.mixin import Injectable

from pokie.contrib.mail.dto import MessageTemplateRecord
from pokie.contrib.mail.helpers import MessageBuilder
from pokie.contrib.mail.repository import MessageTemplateRepository


class MessageTemplateService(Injectable):
    def get_template(
        self, template: str, locale: str, channel: int
    ) -> Optional[MessageTemplateRecord]:
        return self.repo_template.find_template(template, locale, channel)

    def get_builder(
        self, template: str, locale: str, channel: int
    ) -> Optional[MessageBuilder]:
        template = self.repo_template.find_template(template, locale, channel)
        if template is None:
            return None
        return MessageBuilder(template)

    @property
    def repo_template(self) -> MessageTemplateRepository:
        return MessageTemplateRepository(self.get_di().get(DI_DB))



from argparse import ArgumentParser


class ArgParser(ArgumentParser):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.failed = False
        self.error_message = ""

    def error(self, message):
        self.failed = True
        self.error_message = message

    def format_parameters(self):
        formatter = self._get_formatter()

        # positionals, optionals and user-defined groups
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # determine help from format above
        return formatter.format_help()



from typing import List

from rick_db import Repository
from rick_db.sql import Select, Literal, Delete, Insert

from pokie.contrib.auth.dto import AclRole, AclUserRole, AclResource, AclRoleResource


class AclRoleRepository(Repository):
    def __init__(self, db):
        super().__init__(db, AclRole)

    def find_user_roles(self, id_user: int) -> List[AclRole]:
        key = "__acl__:find_user_roles"
        sql = self._cache_get(key)
        if not sql:
            sql, _ = (
                self.select()
                .join(AclUserRole, AclUserRole.id_role, AclRole, AclRole.id)
                .where(AclUserRole.id_user, "=", id_user)
                .order(AclRole.id)
                .assemble()
            )
            self._cache_set(key, sql)
        with self._db.cursor() as c:
            return c.fetchall(sql, [id_user], cls=AclRole)

    def can_remove(self, id_role: int) -> bool:
        """
        Check if a given role is empty or not in use
        :param id_role:
        :return:
        """
        # check if role references resources
        sql, values = (
            Select(self._dialect)
            .from_(
                AclRoleResource,
                cols={Literal("COUNT(*)"): "total"},
                schema=self._schema,
            )
            .where(AclRoleResource.id_role, "=", id_role)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            if result["total"] != 0:
                return False

        # check if role is referenced by users
        sql, values = (
            Select(self._dialect)
            .from_(
                AclUserRole, cols={Literal("COUNT(*)"): "total"}, schema=self._schema
            )
            .where(AclUserRole.id_role, "=", id_role)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            if result["total"] != 0:
                return False

        return True

    def truncate(self, id_role: int):
        # delete role resources
        sql, values = (
            Delete(self._dialect)
            .from_(AclRoleResource)
            .where(AclRoleResource.id_role, "=", id_role)
            .assemble()
        )
        self.exec(sql, values)

        # delete user associations
        sql, values = (
            Delete(self._dialect)
            .from_(AclUserRole)
            .where(AclUserRole.id_role, "=", id_role)
            .assemble()
        )
        self.exec(sql, values)

        # finally, delete role
        self.delete_pk(id_role)

    def add_role_resource(self, id_role: int, id_resource: int):
        """
        Associates a resource with a role
        :param id_role:
        :param id_resource:
        :return:
        """
        # check if resource is already in role
        sql, values = (
            Select(self._dialect)
            .from_(AclRoleResource, schema=self._schema)
            .where(AclRoleResource.id_role, "=", id_role)
            .where(AclRoleResource.id_resource, "=", id_resource)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) > 0:
                return

        sql, values = (
            Insert(self._dialect)
            .into(AclRoleResource(id_role=id_role, id_resource=id_resource))
            .assemble()
        )
        self.exec(sql, values)

    def remove_role_resource(self, id_role: int, id_resource: int):
        """
        Removes an association between a resource and a role
        :param id_role:
        :param id_resource:
        :return:
        """
        # check if resource is already in role
        sql, values = (
            Select(self._dialect)
            .from_(AclRoleResource, schema=self._schema)
            .where(AclRoleResource.id_role, "=", id_role)
            .where(AclRoleResource.id_resource, "=", id_resource)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) == 0:
                return

        # delete resource from role association
        sql, values = (
            Delete(self._dialect)
            .from_(AclRoleResource)
            .where(AclRoleResource.id_role, "=", id_role)
            .where(AclRoleResource.id_resource, "=", id_resource)
            .assemble()
        )
        self.exec(sql, values)

    def remove_user_role(self, id_user: int, id_role: int):
        """
        Removes an association between a user and a role
        :param id_user:
        :param id_role:
        :return:
        """
        # check if role is associated with user
        sql, values = (
            Select(self._dialect)
            .from_(AclUserRole, schema=self._schema)
            .where(AclUserRole.id_role, "=", id_role)
            .where(AclUserRole.id_user, "=", id_user)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) == 0:
                return

        # delete role from user
        sql, values = (
            Delete(self._dialect)
            .from_(AclUserRole)
            .where(AclUserRole.id_role, "=", id_role)
            .where(AclUserRole.id_user, "=", id_user)
            .assemble()
        )
        self.exec(sql, values)

    def add_user_role(self, id_user, id_role):
        """
        Associates a user with a role
        :param id_user:
        :param id_role:
        :return:
        """
        # check if role is associated with user
        sql, values = (
            Select(self._dialect)
            .from_(AclUserRole, schema=self._schema)
            .where(AclUserRole.id_role, "=", id_role)
            .where(AclUserRole.id_user, "=", id_user)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchall(sql, values)
            if len(result) > 0:
                return

        # insert association record
        sql, values = (
            Insert(self._dialect)
            .into(AclUserRole(id_role=id_role, id_user=id_user))
            .assemble()
        )
        self.exec(sql, values)


class AclResourceRepository(Repository):
    def __init__(self, db):
        super().__init__(db, AclResource)

    def find_user_resources(self, id_user: int) -> List[AclResource]:
        key = "__acl__:find_user_resources"
        sql = self._cache_get(key)
        if not sql:
            sql, _ = (
                self.select()
                .join(
                    AclRoleResource,
                    AclRoleResource.id_resource,
                    AclResource,
                    AclResource.id,
                )
                .join(
                    AclUserRole,
                    AclUserRole.id_role,
                    AclRoleResource,
                    AclRoleResource.id_role,
                )
                .where(AclUserRole.id_user, "=", id_user)
                .assemble()
            )
            self._cache_set(key, sql)

        with self._db.cursor() as c:
            return c.fetchall(sql, [id_user], cls=AclResource)

    def find_by_role(self, id_role: int) -> List[AclResource]:
        """
        List resources for a given role
        :param id_role:
        :return:
        """
        key = "__acl__:find_by_role"
        sql = self._cache_get(key)
        if not sql:
            sql, _ = (
                self.select()
                .join(
                    AclRoleResource,
                    AclRoleResource.id_resource,
                    AclResource,
                    AclResource.id,
                )
                .where(AclRoleResource.id_role, "=", id_role)
                .order(AclResource.id)
                .assemble()
            )
            self._cache_set(key, sql)

        with self._db.cursor() as c:
            return c.fetchall(sql, [id_role], cls=AclResource)

    def can_remove(self, id_resource: int):
        """
        Check if a given resource can be removed
        :param id_resource:
        :return:
        """
        # check if role references resources
        sql, values = (
            Select(self._dialect)
            .from_(
                AclRoleResource,
                cols={Literal("COUNT(*)"): "total"},
                schema=self._schema,
            )
            .where(AclRoleResource.id_resource, "=", id_resource)
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchone(sql, values)
            return result["total"] == 0



from typing import List, Optional

from rick_db import Repository
from rick_db.sql import Update
from pokie.contrib.mail.dto import MessageQueueRecord, MessageTemplateRecord
from rick.util.datetime import iso8601_now


class MessageQueueRepository(Repository):
    def __init__(self, db):
        super().__init__(db, MessageQueueRecord)

    def find_by_status(self, channel, status, limit=1000) -> List[MessageQueueRecord]:
        sql, values = (
            self.select()
            .where(MessageQueueRecord.status, "=", status)
            .where(MessageQueueRecord.channel, "=", channel)
            .limit(limit)
        )

        with self._db.cursor() as c:
            return c.fetchall(sql, values, self._record)

    def update_status(self, id: str, status: str):
        record = self.fetch_pk(id)
        if record is not None:
            record.status = status
            if status == "S":
                record.sent = iso8601_now()
            self.update(record)

    def find_first_and_lock(
        self, channel: int, status: str = "Q"
    ) -> Optional[MessageQueueRecord]:
        """
        Locks a queued message and returns the record
        :param status: previous status of record to be locked
        :return:
        """

        qry = (
            self.select(cols=[MessageQueueRecord.id])
            .where(MessageQueueRecord.status, "=", status)
            .where(MessageQueueRecord.channel, "=", channel)
            .limit(1)
            .for_update()
        )

        sql, values = (
            Update(self._dialect)
            .table(self._tablename, self._schema)
            .values({MessageQueueRecord.status: "L"})
            .where(MessageQueueRecord.id, "=", qry)
            .returning()
            .assemble()
        )

        with self._db.cursor() as c:
            result = c.fetchall(sql, values, self._record)
            if len(result) > 0:
                return result[0]
            return None

    def truncate(self):
        sql = "TRUNCATE TABLE {}".format(
            self.dialect().table(self._tablename, schema=self._schema)
        )
        with self._db.cursor() as c:
            c.exec(sql)


class MessageTemplateRepository(Repository):
    def __init__(self, db):
        super().__init__(db, MessageTemplateRecord)

    def find_template(
        self, template: str, language: str, channel: int
    ) -> Optional[MessageTemplateRecord]:
        sql, values = (
            self.select()
            .where(MessageTemplateRecord.template, "=", template)
            .where(MessageTemplateRecord.language, "=", language)
            .where(MessageTemplateRecord.channel, "=", channel)
            .limit(1)
            .assemble()
        )

        with self._db.cursor() as c:
            return c.fetchone(sql, values, self._record)



import signal
from argparse import ArgumentParser
from typing import List

from rick.mixin import Injectable, Runnable
from rick.util.loader import load_class

from pokie.constants import DI_APP, DI_SIGNAL, DI_TTY
from pokie.contrib.base.cli.base import BaseCommand


def abort_jobs(di, signal_no, stack_trace):
    di.get(DI_TTY).write("\nCtrl+C pressed, exiting...")
    exit(0)


class JobBaseCmd(BaseCommand):
    def get_jobs(self) -> dict:
        result = {}
        for module_name, module in self.get_di().get(DI_APP).modules.items():
            jobs = getattr(module, "jobs", [])
            if len(jobs) > 0:
                result[module_name] = jobs
        return result


class JobListCmd(JobBaseCmd):
    description = "list registered job workers"

    def run(self, args) -> bool:
        for name, jobs in self.get_jobs().items():
            self.tty.write("Worker Jobs for module {}:".format(name))
            for job in jobs:
                self.tty.write(
                    self.tty.colorizer.white("   {}".format(job), attr="bold")
                )

        return True


class JobRunCmd(JobBaseCmd):
    description = "run  all job workers"

    def run(self, args) -> bool:
        joblist = []
        di = self.get_di()

        # prepare job list
        for module_name, jobs in self.get_jobs().items():
            for job_name in jobs:
                self.tty.write("Preparing job  '{}'...".format(job_name))
                job = load_class(job_name)
                if job is None:
                    raise ValueError(
                        "Non-existing job class '{}' in module {}".format(
                            job_name, module_name
                        )
                    )
                if not issubclass(job, (Injectable, Runnable)):
                    raise RuntimeError(
                        "Class '{}' must implement Injectable, Runnable interfaces"
                    )
                joblist.append(job(di))

        # run job list
        joblist.reverse()
        self.tty.write("\nRunning jobs, press CTRL+C to abort...")

        # register clean shutdown
        di.get(DI_SIGNAL).add_handler(signal.SIGINT, abort_jobs)

        while True:
            for job in joblist:
                job.run(di)

        # unreachable
        return True



# Version
POKIE_VERSION = ["0", "4", "4"]


def get_version():
    return ".".join(POKIE_VERSION)


# Http Codes
HTTP_OK = 200
HTTP_BADREQ = 400
HTTP_NOAUTH = 401
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND = 404
HTTP_INTERNAL_ERROR = 500

# DI Keys
DI_CONFIG = "config"  # config object
DI_FLASK = "app"  # flask application
DI_APP = "main"  # pokie application
DI_MODULES = "modules"  # module list
DI_SERVICES = "svc_manager"  # service manager
DI_DB = "db"  # database client
DI_REDIS = "redis"  # redis client
DI_EVENTS = "event_manager"  # event manager
DI_TTY = "tty"  # console writer
DI_SIGNAL = "signal"  # signal manager
DI_HTTP_ERROR_HANDLER = "http_error_handler"  # http exception manager

# Error Handler configuration
CFG_HTTP_ERROR_HANLDER = "http_error_handler"

# DB Configuration
CFG_DB_NAME = "db_name"
CFG_DB_HOST = "db_host"
CFG_DB_PORT = "db_port"
CFG_DB_USER = "db_user"
CFG_DB_PASSWORD = "db_password"
CFG_DB_SSL = "db_ssl"

# Redis Configuration
CFG_REDIS_HOST = "redis_host"
CFG_REDIS_PORT = "redis_port"
CFG_REDIS_PASSWORD = "redis_password"
CFG_REDIS_DB = "redis_db"
CFG_REDIS_SSL = "redis_ssl"

# Auth Configuration
CFG_AUTH_SECRET = "auth_secret"

# SMTP Configuration
CFG_SMTP_HOST = "smtp_host"
CFG_SMTP_PORT = "smtp_port"
CFG_SMTP_USE_TLS = "smtp_use_tls"
CFG_SMTP_USE_SSL = "smtp_use_ssl"
CFG_SMTP_DEBUG = "smtp_debug"
CFG_SMTP_USERNAME = "smtp_username"
CFG_SMTP_PASSWORD = "smtp_password"
CFG_SMTP_DEFAULT_SENDER = "smtp_default_sender"
CFG_SMTP_TIMEOUT = "smtp_timeout"
CFG_SMTP_SSL_KEYFILE = "smtp_ssl_keyfile"
CFG_SMTP_SSL_CERTFILE = "smtp_ssl_certfile"



from typing import Any, Optional, List

from rick.base import Di
from rick.mixin import Injectable, Runnable
from rick.util.loader import load_class

from pokie.constants import DI_DB, DI_APP
from pokie.contrib.base.dto import FixtureRecord
from pokie.contrib.base.repository.fixture import FixtureRepository


class FixtureError(Exception):
    pass


class FixtureService(Injectable):
    def __init__(self, di: Di):
        super().__init__(di)
        self.db = di.get(DI_DB)

    def list(self) -> List[FixtureRecord]:
        return self.repo_fixture.fetch_all()

    def add(self, record: FixtureRecord):
        self.repo_fixture.insert_pk(record)

    def scan(self):
        """
        Scans all modules for fixture definitions
        :return: list
        """
        result = []
        app = self.get_di().get(DI_APP)
        for name, module in app.modules.items():
            fixtures = getattr(module, "fixtures", None)
            if fixtures:
                if not isinstance(fixtures, (list, tuple)):
                    raise TypeError(
                        "FixtureManager: invalid data type for fixture list in module '{}'".format(
                            name
                        )
                    )
                result.extend(fixtures)
        return result

    def execute(self, fixture_name: str):
        """
        Attempts to load a fixture and execute it
        :param fixture_name:
        :return:
        """
        cls = load_class(fixture_name)
        if cls is None:
            raise FixtureError(
                "FixtureManager: cannot locate fixture class '{}'".format(fixture_name)
            )

        if not issubclass(cls, (Injectable, Runnable)):
            raise FixtureError(
                "FixtureManager: class '{}' must implement Injectable, Runnable mixins".format(
                    fixture_name
                )
            )

        di = self.get_di()
        fixture = cls(di)
        fixture.run(di)

    @property
    def repo_fixture(self) -> FixtureRepository:
        return FixtureRepository(self.db)



from typing import Optional

from rick.crypto.hasher import HasherInterface
from rick.crypto.hasher.bcrypt import BcryptHasher

from pokie.constants import DI_SERVICES
from pokie.contrib.auth.service.user import AuthUser
from pokie.plugins.auth import AuthPluginInterface
from pokie.contrib.auth.constants import SVC_USER
from pokie.contrib.auth.service import UserService


class DbAuthPlugin(AuthPluginInterface):
    capabilities = [AuthPluginInterface.UPDATE_PASSWORD]

    def autenticate(self, username: str, password: str):
        record = self.svc_user.get_by_username(username)
        if record is None:
            return None

        if not record.active:
            return None

        if self.hasher.is_valid(password, record.password):
            if self.hasher.need_rehash(record.password):
                # update weak password hash
                self.svc_user.update_password(record.id, self.hasher.hash(password))

            # update lastlogin
            self.svc_user.update_lastlogin(record.id)
            return AuthUser(record, self.get_di())

    def load_id(self, id_user) -> Optional[AuthUser]:
        """
        Loads a user record by id
        :param id_user:
        :return:
        """
        record = self.svc_user.get_by_id(id_user)
        if record is None:
            return None

        if not record.active:
            return None

        return AuthUser(record, self.get_di())

    def valid_username(self, username: str) -> bool:
        """
        Checks if the given username is valid (exists and account is enabled)
        :param username:
        :return: True if username exists, false otherwise
        """
        record = self.svc_user.get_by_username(username)
        if record is None:
            return False
        return record.active

    def update_password(self, username: str, password: str) -> bool:
        """
        Updates a user password
        :param username:
        :param password:
        :return:
        """
        if len(password) == 0:
            return False
        record = self.svc_user.get_by_username(username)
        if record is None:
            return False

        self.svc_user.update_password(record.id, self.hasher.hash(password))
        return True

    def is_local(self) -> bool:
        """
        Checks if current backend is local (db/file)
        :return:
        """
        return True

    def has_capability(self, capability: int) -> bool:
        return capability in self.capabilities

    @property
    def svc_user(self) -> UserService:
        return self.get_di().get(DI_SERVICES).get(SVC_USER)

    @property
    def hasher(self) -> HasherInterface:
        return BcryptHasher()



from rick_db import fieldmapper


@fieldmapper(tablename="message_template", pk="id_message_template", schema="public")
class MessageTemplateRecord:
    id = "id_message_template"
    template = "template"
    language = "language"
    created = "created"
    channel = "channel"
    label = "label"
    subject = "subject"
    body = "body"
    html = "html"
    placeholders = "placeholders"


@fieldmapper(tablename="message_queue", pk="id_message_queue", schema="public")
class MessageQueueRecord:
    id = "id_message_queue"
    created = "created"
    sent = "sent"
    channel = "channel"
    status = "status"
    msg_from = "msg_from"
    msg_to = "msg_to"
    title = "title"
    content = "content"
    html = "html"



from rick_db import fieldmapper


@fieldmapper(tablename="user", pk="id_user")
class UserRecord:
    id = "id_user"
    active = "active"
    admin = "admin"
    username = "username"
    first_name = "first_name"
    last_name = "last_name"
    email = "email"
    creation_date = "creation_date"
    last_login = "last_login"
    password = "password"
    external = "external"
    attributes = "attributes"



import logging

from rick.base import Di
from rick.mixin import Injectable, Runnable
from rick.resource.console import ConsoleWriter
from rick_mailer import Mailer, SMTPFactory

from pokie.constants import DI_TTY, DI_SERVICES, DI_CONFIG
from pokie.contrib.mail.constants import SVC_MESSAGE_QUEUE
from pokie.contrib.mail.service import MessageQueueService


class MailJob(Injectable, Runnable):
    def __init__(self, di: Di):
        super().__init__(di)
        self.tty = di.get(DI_TTY) if di.has(DI_TTY) else ConsoleWriter()

    def get_smtp_mailer(self) -> Mailer:
        config = self.get_di().get(DI_CONFIG).asdict()
        return Mailer(SMTPFactory(config))


class MailQueueJob(MailJob):
    MAILS_PER_RUN = 10000

    def run_smtp(self):
        total = 0
        mailer = self.get_smtp_mailer()
        while total < self.MAILS_PER_RUN:
            record = self.svc_queue.fetch_for_processing(
                MessageQueueService.CHANNEL_SMTP
            )
            if record is None:
                # no more messages, exit
                return True

            try:
                if (
                    mailer.send_mail(
                        record.title,
                        record.content,
                        record.msg_from,
                        record.msg_to.split(","),
                        html_message=record.html,
                    )
                    > 0
                ):
                    self.svc_queue.update_status(
                        record.id, MessageQueueService.STATUS_SENT
                    )
            except Exception as e:
                raise e
                self.tty.error(str(e))
                self.svc_queue.update_status(
                    record.id, MessageQueueService.STATUS_FAILED
                )
                return False
        return True

    def run(self, di: Di):
        # run smtp - channel 0
        return self.run_smtp()

    @property
    def svc_queue(self) -> MessageQueueService:
        return self.get_di().get(DI_SERVICES).get(SVC_MESSAGE_QUEUE)



class ParseListError(Exception):
    pass


def parse_list_parameters(args, record):
    search_text = args.get("search")
    match_fields = args.get("match")
    offset = args.get("offset")
    limit = args.get("limit")
    sort = args.get("sort")

    # match fields
    if match_fields is not None:
        match_fields = match_fields.split("|")
        # convert field names to column names
        result = {}
        for f in match_fields:
            f = f.split(":", 1)
            if len(f) != 2:
                raise ParseListError("invalid field match expression")
            name = f[0]
            if name not in record._fieldmap.keys():
                raise ParseListError("invalid field name: {}".format(f))
            result[record._fieldmap[name]] = f[1]
        # replace original dict with result
        match_fields = result

    if offset is not None:
        try:
            offset = int(offset)
            if offset < 0:
                raise ValueError
        except ValueError:
            raise ParseListError("invalid offset value")

    if limit is not None:
        try:
            limit = int(limit)
            if limit < 1:
                raise ValueError
        except ValueError:
            raise ParseListError("invalid limit value")

    # validate sort fields
    if sort is not None:
        sort = sort.split(",")
        result = {}
        # convert field, field:desc -> db_field:asc, db_field:desc
        for expr in sort:
            expr = expr.split(":")
            if expr[0] not in record._fieldmap.keys():
                raise ParseListError("invalid sort field name: {}".format(expr[0]))
            name = record._fieldmap[expr[0]]

            if len(expr) > 1:
                if expr[1].lower() not in ["asc", "desc"]:
                    raise ParseListError("invalid sort order: {}".format(expr[1]))
                result[name] = expr[1]
            else:
                result[name] = "asc"
        sort = result

    return search_text, match_fields, limit, offset, sort



from rick.base import Di
from rick.mixin import Translator
from rick.validator import registry
from rick.validator import Rule

from pokie.constants import DI_SERVICES
from pokie.contrib.base.constants import SVC_VALIDATOR

# dependency injector
_di = None


def init_validators(di: Di):
    global _di
    _di = di


@registry.register_cls(name="pk")
class DbPrimaryKey(Rule):
    MSG_ERROR = "Invalid primary key value"

    def validate(
        self, value, options: list = None, error_msg=None, translator: Translator = None
    ):
        if len(options) == 0:
            raise RuntimeError("DbPrimaryKey(): missing table name")

        if _di is None:
            raise RuntimeError("DbPrimaryKey(): di not initialized")

        table_name = str(options[0])
        tokens = table_name.split(".")
        if len(tokens) > 2 or len(tokens) == 0:
            raise ValueError(
                "DbPrimaryKey(): invalid table name '{}'".format(table_name)
            )
        schema = None
        table_name = tokens[0]
        if len(tokens) == 2:
            schema = tokens[1]

        pk_name = ""
        if len(options) > 1:
            pk_name = str(options[1])

        svc = _di.get(DI_SERVICES).get(SVC_VALIDATOR)
        try:
            if svc.id_exists(pk_name, value, table_name, schema):
                return True, ""
            return False, self.error_message(error_msg, translator)

        except Exception:
            return False, self.error_message(error_msg, translator)



import abc

from rick.mixin import Injectable


class BaseModule(Injectable, abc.ABC):
    # base module name; must be unique
    name = ""

    # module extended description
    description = ""

    # service mapper
    services = {}

    # cli command mapper
    cmd = {}

    # jobs
    jobs = []

    def build(self, parent=None):
        """
        Initialize module internals
        :param parent: FlaskApplication instance
        :return:
        """
        pass



import inspect
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional, List

from rick_db.conn import Connection
from rick_db.sql.dialect import PgSqlDialect
from rick_db.util.pg import PgInfo

from pokie.codegen.pg import PgTableSpec, RecordGenerator, RequestGenerator
from pokie.constants import DI_DB
from pokie.core import CliCommand


class DbCodeGenCommand(CliCommand):
    def get_db(self) -> Optional[Connection]:
        result = None
        di = self.get_di()
        if di.has(DI_DB):
            result = di.get(DI_DB)
        return result

    def is_supported(self, db) -> bool:
        if isinstance(db, Connection):
            return isinstance(db.dialect(), PgSqlDialect)
        return False

    def parse_table_list(self, mgr: PgInfo, table_expr) -> [List, str]:
        schema = mgr.SCHEMA_DEFAULT
        table_list = []
        table_expr = table_expr.split(".", 1)
        table = table_expr[0]
        if len(table_expr) > 1:
            schema = table_expr[0]
            table = table_expr[1]

        all_tables = []
        for tbl in mgr.list_database_tables(schema):
            all_tables.append(tbl.name)

        if table == "*":
            table_list = all_tables
        else:
            if table not in all_tables:
                self.tty.error(
                    "table with name '{}' not found in schema '{}'".format(
                        table, schema
                    )
                )
                return []
            table_list.append(table)

        return table_list, schema

    def pg_gen_dto(self, db, table_expr, dest_file, camel_case=False) -> bool:
        pg = PgTableSpec(db)
        table_list, schema = self.parse_table_list(pg.manager(), table_expr)
        if len(table_list) == 0:
            return False

        gen = RecordGenerator()
        first = True
        result = []
        for name in table_list:
            self.tty.write(
                self.tty.colorizer.white(
                    "generating dto for {}.{}...".format(schema, name), attr="bold"
                )
            )
            spec = pg.table_spec(name, schema)
            result.append(gen.gen_source(spec, camelcase=camel_case, imports=first))
            first = False

        if dest_file is None:
            self.tty.write("\n".join(result))
        else:
            self.tty.write(
                self.tty.colorizer.white("writing to file '{}'...".format(dest_file))
            )
            with open(dest_file, "w") as f:
                f.write("\n".join(result))

            self.tty.write(self.tty.colorizer.green("success!"))
        return True

    def pg_gen_request(
        self, db, table_expr, dest_file, camelcase_id=False, camelcase_cols=False
    ) -> bool:
        pg = PgTableSpec(db)
        table_list, schema = self.parse_table_list(pg.manager(), table_expr)
        if len(table_list) == 0:
            return False

        gen = RequestGenerator()
        first = True
        result = []
        for name in table_list:
            self.tty.write(
                self.tty.colorizer.white(
                    "generating RequestRecord class for {}.{}...".format(schema, name),
                    attr="bold",
                )
            )
            spec = pg.table_spec(name, schema)
            result.append(
                gen.gen_source(
                    spec,
                    camelcase=camelcase_id,
                    db_camelcase=camelcase_cols,
                    imports=first,
                )
            )
            first = False

        if dest_file is None:
            self.tty.write("\n".join(result))
        else:
            self.tty.write(
                self.tty.colorizer.white("writing to file '{}'...".format(dest_file))
            )
            with open(dest_file, "w") as f:
                f.write("\n".join(result))

            self.tty.write(self.tty.colorizer.green("success!"))
        return True


class GenDtoCmd(DbCodeGenCommand):
    description = "generate dto class from database table"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "table", type=str, help="source table or view for dto generation"
        )
        parser.add_argument("-f", dest="file", type=str, help="destination file")
        parser.add_argument(
            "-c",
            "--camelcase",
            action="store_true",
            default=False,
            help="camelCase attributes",
        )

    def run(self, args) -> bool:
        db = self.get_db()
        if not self.is_supported(db):
            self.tty.error("database code generation is only supported with PostgreSQL")
            return False

        if args.file is not None:
            dest_file = Path(args.file)
            if dest_file.exists():
                self.tty.error("destination file already exists")
                return False

        return self.pg_gen_dto(db, args.table, args.file, args.camelcase)


class GenRequestRecordCmd(DbCodeGenCommand):
    description = "generate RequestRecord class from database table"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "table", type=str, help="source table or view for request record generation"
        )
        parser.add_argument("-f", dest="file", type=str, help="destination file")
        parser.add_argument(
            "-c",
            "--camelcase-names",
            action="store_true",
            default=False,
            help="camelCase field names",
        )
        parser.add_argument(
            "-C",
            "--camelcase-cols",
            action="store_true",
            default=False,
            help="camelCase bind columns",
        )

    def run(self, args) -> bool:
        db = self.get_db()
        if not self.is_supported(db):
            self.tty.error("database code generation is only supported with PostgreSQL")
            return False

        if args.file is not None:
            dest_file = Path(args.file)
            if dest_file.exists():
                self.tty.error("destination file already exists")
                return False

        return self.pg_gen_request(
            db, args.table, args.file, args.camelcase_names, args.camelcase_cols
        )



import json
import re

from pokie.contrib.mail.dto import MessageTemplateRecord, MessageQueueRecord
from pokie.contrib.mail.service import MessageQueueService


class MessageBuilder:
    def __init__(self, template: MessageTemplateRecord):
        self.template = template

    def assemble(self, msg_from, msg_to, data: dict):
        message = MessageQueueRecord(
            channel=self.template.channel, status=MessageQueueService.STATUS_QUEUED
        )
        if len(self.template.placeholders) > 0:
            tmp = json.loads(self.template.placeholders)
            data = {**tmp, **data}

        if isinstance(msg_to, (list, tuple)):
            msg_to = ",".join(msg_to)

        if not msg_from or not msg_to:
            raise ValueError("assemble(): msg_from and msg_to cannot be empty")

        message.msg_from = msg_from
        message.msg_to = msg_to
        message.title = self._replace(data, self.template.subject)
        message.content = (
            self._replace(data, self.template.body)
            if len(self.template.body) > 0
            else ""
        )
        message.html = (
            self._replace(data, self.template.html)
            if len(self.template.html) > 0
            else None
        )
        return message

    def _replace(self, replace_map: dict, src) -> str:
        regexp = re.compile("|".join(map(re.escape, replace_map)))
        return regexp.sub(lambda match: replace_map[match.group(0)], src)



from rick.util.string import snake_to_camel, snake_to_pascal
from rick.form import RequestRecord, field
from pokie.codegen.spec import FieldSpec, TableSpec
from pokie.codegen.textfile import TextBuffer


class RequestGenerator:
    # default validators for misc data types
    validators = {
        "int2": ["numeric"],
        "int4": ["numeric"],
        "int8": ["numeric"],
        "numeric": ["decimal"],
        "bool": ["bool"],
        "timestamptz": ["iso8601"],
        "timestamp": ["iso8601"],
        "date": ["iso8601"],
    }

    def _field(self, f: FieldSpec, camelcase=False, db_camelcase=False) -> tuple:
        """
        Generate a RequestRecord field definition
        :param f:
        :param camelcase:
        :param db_camelcase:
        :return: touple with (name, validators, bind_name)
        """
        validators = []
        if not f.pk:
            name = f.name if camelcase is False else snake_to_camel(f.name)
            target = f.name if db_camelcase is False else snake_to_camel(f.name)

            if not f.nullable:
                validators.append("required")
        else:
            # primary key name is always id
            # if it is an auto number, validate as id
            name = "id"
            target = "id"
            if f.auto:
                validators.append("id")

        # add predefined validators for data types
        if f.dtype in self.validators.keys():
            for v in self.validators[f.dtype]:
                validators.append(v)

        # add maxlen if defined in spec
        if "maxlen" in f.dtype_spec.keys():
            validators.append("maxlen:{}".format(str(f.dtype_spec["maxlen"])))

        # add foreign key lookup
        if f.fk:
            validators.append(
                "pk:{}.{},{}".format(f.fk_schema, f.fk_table, f.fk_column)
            )

        return name, validators, target

    def _gen_field_src(self, f: FieldSpec, camelcase=False, db_camelcase=False) -> str:
        """
        Assemble a field definition source code line
        :param f:
        :param camelcase:
        :param db_camelcase:
        :return: str
        """
        name, validators, target = self._field(f, camelcase, db_camelcase)
        validators = "|".join(validators)
        return "'{name}': field(validators='{validators}', bind='{bind}'),".format(
            name=name, validators=validators, bind=target
        )

    def gen_source(
        self,
        spec: TableSpec,
        camelcase=False,
        db_camelcase=False,
        gen: TextBuffer = None,
        imports=True,
    ):
        """
        Generate a Rick RequestRecord source file

        :param spec: Table spec
        :param camelcase: if True, attributes will be camelCased
        :param db_camelcase: if True, db record attributes will be camelCased
        :param gen: optional TextBuffer
        :param imports: if True, generate import line
        :return: source code string
        """
        if gen is None:
            gen = TextBuffer()

        if imports is True:
            gen.writeln("from rick.form import RequestRecord, field", newlines=2)

        gen.writeln(
            "class {}Request(RequestRecord):".format(snake_to_pascal(spec.table))
        )
        gen.writeln("fields = {", level=1)
        for f in spec.fields:
            gen.writeln(self._gen_field_src(f, camelcase, db_camelcase), level=2)
        gen.writeln("}", level=1)
        gen.writeln(gen.newline())
        return gen.read()

    def gen_class(
        self, spec: TableSpec, camelcase=False, db_camelcase=False
    ) -> RequestRecord:
        """
        Generate a Rick RequestRecord class from a spec
        :param spec:
        :param camelcase: if True, attributes will be camelCased
        :param db_camelcase: if True, db record attributes will be camelCased
        :return: RequestRecord class
        """

        class reqRecord(RequestRecord):
            pass

        fields = {}
        for f in spec.fields:
            name, validators, target = self._field(f, camelcase, db_camelcase)
            fields[name] = field(validators="|".join(validators), bind=target)

        setattr(reqRecord, "fields", fields)

        return reqRecord



from rick.base import Di
from rick.mixin import Injectable
from rick_db import Repository

from pokie.constants import DI_DB


class AutoRestService(Injectable):
    def __init__(self, di: Di):
        super().__init__(di)
        self.repositories = {}

    def get_record(self, class_record, id_record):
        return self.get_repository(class_record).fetch_pk(id_record)

    def delete_record(self, class_record, id_record):
        return self.get_repository(class_record).delete_pk(id_record)

    def insert_record(self, class_record, record):
        return self.get_repository(class_record).insert_pk(record)

    def update_record(self, class_record, record):
        return self.get_repository(class_record).update(record)

    def get_repository(self, class_record) -> Repository:
        if class_record not in self.repositories.keys():
            self.repositories[class_record] = Repository(
                self.get_di().get(DI_DB), class_record
            )
        return self.repositories[class_record]



from rick.base import Di

from pokie.constants import DI_DB
from rick.mixin import Injectable
from rick_db import DbGrid, Repository
from .mixin import RestServiceMixin


class RestService(Injectable, RestServiceMixin):
    record_class = None  # record class
    repository_class = None  # optional custom repository class

    def __init__(self, di: Di):
        super().__init__(di)
        # copy class-attributes to instance-attributes
        # this allows usage of multiple instances of RestService with different record and repository classes
        self._record_cls = self.record_class
        self._repository_cls = self.repository_class

    def set_record_class(self, cls):
        self._record_cls = cls

    def set_repository_class(self, cls):
        self._repository_cls = cls

    def get(self, id_record):
        return self.repository.fetch_pk(id_record)

    def delete(self, id_record):
        return self.repository.delete_pk(id_record)

    def insert(self, record):
        return self.repository.insert_pk(record)

    def update(self, id_record, record):
        return self.repository.update(record, id_record)

    def exists(self, id_record):
        return self.repository.valid_pk(id_record)

    def list(
        self,
        search_fields: list,
        search_text: str = None,
        match_fields: dict = None,
        limit: int = None,
        offset: int = None,
        sort_fields: dict = None,
        search_filter: list = None,
    ):
        grid = DbGrid(self.repository, search_fields, DbGrid.SEARCH_ANY)
        return grid.run(
            None,
            search_text=search_text,
            match_fields=match_fields,
            limit=limit,
            offset=offset,
            sort_fields=sort_fields,
            search_fields=search_filter,
        )

    @property
    def repository(self) -> Repository:
        if self._record_cls is None:
            raise RuntimeError("Missing record class for repository")
        if self._repository_cls is None:
            return Repository(self.get_di().get(DI_DB), self._record_cls)
        else:
            return self._repository_cls(self.get_di().get(DI_DB))



from abc import ABC
from argparse import ArgumentParser

from rick.base import Di
from rick.resource.console import ConsoleWriter
from rick.mixin import Injectable

from pokie.constants import DI_TTY


class CliCommand(ABC, Injectable):
    description = "command description"
    skipargs = False

    def __init__(self, di: Di, writer=None):
        self.set_di(di)
        if not writer:
            if di.has(DI_TTY):
                writer = di.get(DI_TTY)
            else:
                writer = ConsoleWriter()
        self.tty = writer

    def arguments(self, parser: ArgumentParser):
        pass

    def run(self, args) -> bool:
        pass



import os

import pytest
import sys
from pokie.core import CliCommand


class PyTestCmd(CliCommand):
    description = "run pytest"
    skipargs = True

    def run(self, args) -> bool:
        args = []
        if len(sys.argv) > 2:
            args = sys.argv[2:]
        self.tty.write(
            self.tty.colorizer.white("[Pokie]", attr="bold")
            + " Running pytest with: {}".format(str(args))
        )
        sys.exit(pytest.main(args, plugins=["pytest_pokie"]))
